/******************************************************************************
* augmentation.h
*
* RouteSketch
*
******************************************************************************
* Copyright (C) 2013 Andreas Gemsa <gemsa@kit.edu>, Thomas Pajor
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation, either version 3 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/
/*
 * augmentation.h
 *
 *  Created on: Jul 7, 2010
 *      Author: pajor
 */

#ifndef AUGMENTATION_H_
#define AUGMENTATION_H_

#include "../../../dataStructure/staticTimeGraph.h"
#include "../dataStructures/graph.h"
#include "simplification.h"
#include "../geometry/tools.h"
#include "../geometry/latLng.h"
#include "../geometry/point.h"
#include "../geometry/mercatorProjection.h"

#include <set>

#include <boost/foreach.hpp>
#include <boost/graph/graph_traits.hpp>

namespace schematization {

// This augments the graph with context information
template<typename graphType, typename outGraphType>
void augment_graph(graphType &algo_graph, streetNamesType &street_names, outGraphType &out_graph, set<NodeID>& forbidden_nodes) {
	typedef typename outGraphType::vertex_descriptor vertex_descriptor;

	// Iterate over all vertices and add context-edges to vertices which are important
	// and are "inner vertices", i.e., of degree > 1
	size_t num(0);
	BOOST_FOREACH(vertex_descriptor v, vertices(out_graph)) {
		if (!is_important_node(v, out_graph))
			continue;
		// Skip vertices with degree < 2
		if (in_degree(v, out_graph) + out_degree(v, out_graph) < 2)
			continue;

		add_incident_streets_to_node(algo_graph, street_names, v, out_graph, forbidden_nodes);
//		++num;
//		if (num == 1) break;
	}
}


// This adds incident streets to a node
template<typename graphType, typename outGraphType>
void add_incident_streets_to_node(graphType &algo_graph, streetNamesType &street_names, typename outGraphType::vertex_descriptor v, outGraphType &out_graph, set<NodeID>& forbidden_nodes) {
	// This set contains nodes (neighbors) to which NO edges should be added
	// i.e., nodes that are already contained in out_graph
	set<NodeID> banned_nodes;

	set<NodeID> new_nodes;

	typedef typename outGraphType::vertex_descriptor vertex_descriptor;
	BOOST_FOREACH(vertex_descriptor w, adjacent_vertices(v, out_graph)) {
		banned_nodes.insert(out_graph[w].node_id);
	}
	BOOST_FOREACH(vertex_descriptor w, inv_adjacent_vertices(v, out_graph)) {
		banned_nodes.insert(out_graph[w].node_id);
	}

	// Iterate over all out-edges and add them to t3h graph.
	forall_out_edges(algo_graph, e, out_graph[v].node_id) {
		NodeID t = algo_graph.edges[e].targetNode;

		// If the target node t is banned, skip it
		if (banned_nodes.find(t) != banned_nodes.end() || forbidden_nodes.find(t) != forbidden_nodes.end())
			continue;

		vertex_descriptor new_node = add_vertex(out_graph);
		out_graph[new_node].node_id = t;
		out_graph[new_node].coord = mercatorProjection::latlng_to_point(latLng(algo_graph.nodes[t].data.y, algo_graph.nodes[t].data.x));
		out_graph[new_node].lat_lng = latLng(algo_graph.nodes[t].data.y, algo_graph.nodes[t].data.x);

		// Add an edge to the just created node
		typename outGraphType::edge_descriptor new_edge;
		bool edge_created;
		tie(new_edge, edge_created) = add_edge(v, new_node, out_graph);
		out_graph[new_edge].category = algo_graph.edges[e].data.category;
		out_graph[new_edge].name = "";//street_names[algo_graph.edges[e].data.street_name_id];
		out_graph[new_edge].context = true;
	}

}

}

#endif /* AUGMENTATION_H_ */
