/******************************************************************************
* merge_graphs.h
*
* RouteSketch
*
******************************************************************************
* Copyright (C) 2013 Andreas Gemsa <gemsa@kit.edu>, Thomas Pajor
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation, either version 3 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/
/*
 * merge_graphs.h
 *
 *  Created on: Jul 7, 2010
 *      Author: pajor
 */

#ifndef MERGE_GRAPHS_H_
#define MERGE_GRAPHS_H_

#include "../../../dataStructure/staticTimeGraph.h"
#include "../dataStructures/graph.h"
#include "../geometry/tools.h"
#include "../geometry/line.h"


#include <boost/graph/iteration_macros.hpp>
#include <boost/graph/graph_traits.hpp>

#include <set>
#include <map>

namespace schematization {

template<typename schemGraphType>
void merge_graphs( vector<schemGraphType>& graphs, schemGraphType &out_graph) {
	typedef typename schemGraphType::vertex_descriptor vertex_descriptor;
	typedef typename schemGraphType::edge_descriptor edge_descriptor;

	
	map <NodeID, vertex_descriptor> vertex_map;

	for (size_t i = 0; i < graphs.size(); i++){
		vertex_descriptor v;

		BGL_FORALL_VERTICES_T(v, graphs[i], schemGraphType){

			if (vertex_map.find(graphs[i][v].node_id) == vertex_map.end()) {

				vertex_descriptor new_node = add_vertex(out_graph);
				out_graph[new_node].node_id = graphs[i][v].node_id;
				out_graph[new_node].coord = graphs[i][v].coord;
				
				out_graph[new_node].lat_lng = graphs[i][v].lat_lng;
				
				vertex_map[graphs[i][v].node_id] = new_node;
			}
		}

	}


	
	typedef std::pair<NodeID, NodeID> vertex_pair;
	map <vertex_pair, bool> edge_map;

	for (size_t i = 0; i < graphs.size(); i++){
		edge_descriptor e;
		BGL_FORALL_EDGES_T(e, graphs[i], schemGraphType){
			vertex_descriptor s = source(e, graphs[i]);
			vertex_descriptor t = target(e, graphs[i]);	
			NodeID s_id = graphs[i][s].node_id;
			NodeID t_id = graphs[i][t].node_id;
			vertex_pair p(s_id, t_id);
			vertex_pair p2(t_id, s_id);
			
			if ( (edge_map.find(p) == edge_map.end()) && (edge_map.find(p2) == edge_map.end())) {

				bool edge_created;
				edge_descriptor new_edge;				
				tie(new_edge, edge_created) = add_edge(vertex_map[s_id], vertex_map[t_id], out_graph);

				out_graph[new_edge] = graphs[i][e];
				edge_map[p] = true;
			
			} 
			
		}

	}

	cout << "Merged Graph: " << num_vertices(out_graph) << " vertices and " << num_edges(out_graph) << " edges" << endl;

	//looking for intersections
	
}

template<typename schemGraphType>
bool has_intersection(const schemGraphType &graph){
	typedef typename schemGraphType::vertex_descriptor vertex_descriptor;
	typedef typename schemGraphType::edge_descriptor edge_descriptor;
	typedef typename schemGraphType::edge_iterator edge_iterator;
	edge_iterator e_it_outer, e_it_inner, last_e_it;

	for (tie(e_it_outer, last_e_it) = edges(graph); e_it_outer != last_e_it; ++e_it_outer) {

		vertex_descriptor p = source(*e_it_outer, graph);
		vertex_descriptor q = target(*e_it_outer, graph);
		
		line l1 = line(point(graph[p].coord.x, graph[p].coord.y), point(graph[q].coord.x, graph[q].coord.y) );

		for ( e_it_inner = e_it_outer; e_it_inner != last_e_it; ++e_it_inner) {
			if (e_it_inner == e_it_outer)
					continue;
			vertex_descriptor r = source(*e_it_inner, graph);
			vertex_descriptor s = target(*e_it_inner, graph);

			line l2 = line(point(graph[r].coord.x, graph[r].coord.y), point(graph[s].coord.x, graph[s].coord.y) );
			
			if (tools::is_intersecting(l1, l2))
				cout << "schnitt!" << *e_it_outer << " " << *e_it_inner << endl;

			
		}
	}
	
	
}

}

#endif /* MERGE_GRAPHS_H_ */
