/******************************************************************************
* replace_intersections.h
*
* RouteSketch
*
******************************************************************************
* Copyright (C) 2013 Andreas Gemsa <gemsa@kit.edu>, Thomas Pajor
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation, either version 3 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/
/*
 * bypass_intersections.h
 *
 *  Created on: Jul 1, 2010
 *      Author: gemsa
 */

#ifndef REPLACE_INTERSECTIONS_H_
#define REPLACE_INTERSECTIONS_H_

#include "../dataStructures/graph.h"
#include "../geometry/tools.h"
#include "../geometry/mercatorProjection.h"

#include <vector>

namespace schematization {

template<typename graphType>
class replace_intersections {
public:
    replace_intersections(graphType& input_graph) : graph(input_graph)
    {
        
    }

	size_t run() {
	/*
		detect already existing intersections
		*/
		bool found_intersection = true;
		while (found_intersection){
			found_intersection = false;	

            edge_descriptor e1;
			BGL_FORALL_EDGES_T(e1, graph, graphType){
				if (graph[e1].context) continue;
				
				//cout << "edge e1: " << e1 << endl;

                edge_descriptor e2;
				BGL_FORALL_EDGES_T(e2, graph, graphType){
					if (graph[e2].context) continue;
					
					if (e1 == e2) continue;					

    				//cout << "edge e2: " << e2 << endl;
    
					double x, y;
					bool intersection = test_for_intersection(e1, e2, x, y);


					if (intersection)
					{
						cout << "found intersection: " << e1 << " " << e2 << endl;
						cout << x << ", " << y << endl;
						
						   
						vertex_descriptor new_node;
						edge_descriptor e1_a, e1_b, e2_a, e2_b;

						vertex_descriptor u(source(e1, graph));
						vertex_descriptor v(target(e1, graph));
						vertex_descriptor w(source(e2, graph));
						vertex_descriptor z(target(e2, graph));
                        
                        /*
						cout << "u: " << graph[u].coord.x << ", " << graph[u].coord.y << endl;
						cout << "v: " << graph[v].coord.x << ", " << graph[v].coord.y << endl;
						cout << "w: " << graph[w].coord.x << ", " << graph[w].coord.y << endl;
						cout << "z: " << graph[z].coord.x << ", " << graph[z].coord.y << endl;
                        
                        char c;
                        cin >> c;
                        */
                            
						new_node = add_vertex(graph);

						//new node
						graph[new_node].coord = point(x, y);
                        graph[new_node].lat_lng = mercatorProjection::point_to_latlng(graph[new_node].coord);
						graph[new_node].node_id =  DUMMY_NODE_ID;

						int cat_e1 = graph[e1].category;
						int cat_e2 = graph[e2].category;
						string name_e1 = graph[e1].name;
						string name_e2 = graph[e2].name;
						
						//new edges
						bool edge_added;
						tie (e1_a, edge_added) = add_edge(u, new_node, graph);
						graph[e1_a].category = cat_e1;
						graph[e1_a].name = name_e1;

						tie (e1_b, edge_added) = add_edge(new_node, v, graph);
						graph[e1_b].category = cat_e1;
						graph[e1_b].name = name_e1;

						tie (e2_a, edge_added) = add_edge(w, new_node, graph);
						graph[e2_a].category = cat_e2;
						graph[e2_a].name = name_e2;

						tie (e2_b, edge_added) = add_edge(new_node, z, graph);
						graph[e2_b].category = cat_e2;
						graph[e2_b].name = name_e2;
						
						
						remove_edge(e1, graph);
						remove_edge(e2, graph);						
					
						found_intersection = true;
						break;
						
					}
				}
				if (found_intersection)
					break;
			}
		}

	}

    bool test_for_intersection(const typename graphType::edge_descriptor e1, const typename graphType::edge_descriptor e2, double& x, double& y)
	{
		// Edges which are adjacent (or the same) are never conflicting...
		if (e1 == e2 || is_adjacent(e1, e2, graph))
			return false;
		
		// We say e1 = (u1, v1) and e2 = (u2, v2)
		// Thus we have [(xu1,yu1),(xv1,yv1)] and [(xu2,yu2),(xv2,yv2)]
		vertex_descriptor u1(source(e1, graph)), v1(target(e1, graph)), u2(source(e2, graph)), v2(target(e2, graph));

		const double xu1 = graph[u1].coord.x;
		const double yu1 = graph[u1].coord.y;
		const double xu2 = graph[u2].coord.x;
		const double yu2 = graph[u2].coord.y;
		const double xv1 = graph[v1].coord.x;
		const double yv1 = graph[v1].coord.y;
		const double xv2 = graph[v2].coord.x;
		const double yv2 = graph[v2].coord.y;
		
        double eps = 0.0001;
        if (abs(xu1 - xu2) < eps && abs(yu1 - yu2) < eps)
            return false;
        if (abs(xu1 - xv1) < eps && abs(yu1 - yv1) < eps)
            return false;
        if (abs(xu1 - xv2) < eps && abs(yu1 - yv2) < eps)
            return false;
        if (abs(xu2 - xv1) < eps && abs(yu2 - yv1) < eps)
            return false;
        if (abs(xu2 - xv2) < eps && abs(yu2 - yv2) < eps)
            return false;
        if (abs(xv1 - xv2) < eps && abs(yv1 - yv2) < eps)
            return false;

		const double A1 = yv1 - yu1;
		const double B1 = xu1 - xv1;
		const double C1 = A1*xu1 + B1*yu1;
		const double A2 = yv2 - yu2;
		const double B2 = xu2 - xv2;
		const double C2 = A2*xu2 + B2*yu2;

		const double det = A1*B2 - A2*B1;

		bool intersect = true;
        
        
        //cout << xu1 << ", " << yu1 << " - " << xv1 << ", " << yv1 << endl;
        //cout << xu2 << ", " << yu2 << " - " << xv2 << ", " << yv2 << endl;

		// Lines are parallel, so no overlap
		if (det == 0)
			intersect = false;
		else {
			// If line Segments are not parallel then test for intersection...
			// Compute intersection point
			x = (B2*C1 - B1*C2)/det;
			y = (A1*C2 - A2*C1)/det;
			
		if ( (x == xu1 && y == yu1) ||
             (x == xu2 && y == yu2) ||
             (x == xv1 && y == yv1) ||
             (x == xv2 && y == yv2))
    		return false;

			// Test whether intersection point is contained on the line segments
			double epsilon = 0.0001; //circumenvents problems with float point arithmetic
			if (! (std::min(xu1,xv1) - epsilon <= x && x <= std::max(xu1,xv1) + epsilon) )
				intersect = false;
			if (! (std::min(yu1,yv1) - epsilon <= y && y <= std::max(yu1,yv1) + epsilon) )
				intersect = false;
			if (! (std::min(xu2,xv2) - epsilon <= x && x <= std::max(xu2,xv2) + epsilon) )
				intersect = false;
			if (! (std::min(yu2,yv2) - epsilon <= y && y <= std::max(yu2,yv2) + epsilon) )
				intersect = false;
		}
		return intersect;
	}

private:
    graphType &graph;

private:
	typedef typename graphType::vertex_descriptor vertex_descriptor;
	typedef typename graphType::edge_descriptor edge_descriptor;
	typedef typename graph_traits<graphType>::vertex_iterator vertex_iterator;
	typedef typename graph_traits<graphType>::edge_iterator edge_iterator;
};

}

#endif /* REPLACE_INTERSECTIONS_H_ */
