/******************************************************************************
* simplification.h
*
* RouteSketch
*
******************************************************************************
* Copyright (C) 2013 Andreas Gemsa <gemsa@kit.edu>, Thomas Pajor
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation, either version 3 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/
/*
 * simplification.h
 *
 *  Created on: Jul 1, 2010
 *      Author: pajor
 */

#ifndef SIMPLIFICATION_H_
#define SIMPLIFICATION_H_

#include "../dataStructures/graph.h"
#include "../dataStructures/path.h"

#include "../algorithms/path_extraction.h"
#include "../algorithms/pathDFS.h"
#include "../algorithms/ramerDouglasPeucker.h"
#include "../algorithms/bypass_intersections.h"

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/iteration_macros.hpp>

#define DEBUG_LINE std::cout << __PRETTY_FUNCTION__ << ", line " << __LINE__ << std::endl;

namespace schematization {

// This algorithm simplifies all paths in the graph
template<typename graphType>
void simplify_graph(graphType &in_graph, graphType &out_graph, const double eps) {

	// Clear out graph
	out_graph.clear();

	// This is a vertex map, which maps vertices in the old graph, to their counterparts in the new graph.
	// Note, only "important" vertices (i.e., not being an inner path node) are mapped.
	typedef std::map<typename graphType::vertex_descriptor, typename graphType::vertex_descriptor> vertexMapType;
	vertexMapType vertex_map;


	// Iterate over all nodes and mark important nodes that must not be removed.
	// Moreover, add the node to the out graph and map its id.
	typedef typename graph_traits<graphType>::vertex_iterator vertex_iterator;
	vertex_iterator current_v, last_v;
	for (tie(current_v, last_v) = vertices(in_graph); current_v != last_v; ++current_v) {
		in_graph[*current_v].important = is_important_node(*current_v, in_graph);
		if (in_graph[*current_v].important) {
			typename graphType::vertex_descriptor new_node = add_vertex(out_graph);
			out_graph[new_node] = in_graph[*current_v]; // copy node properties.
			vertex_map[*current_v] = new_node;
		}
	}
    

	// Iterate over all nodes, and from every important node v, start a DFS for every
	// outgoing edge from v.
	pathDFS<graphType> dfs(in_graph);

	for (tie(current_v, last_v) = vertices(in_graph); current_v != last_v; ++current_v) {
		if (!in_graph[*current_v].important)
			continue;

		// Get outgoing edges
		typedef typename graph_traits<graphType>::out_edge_iterator out_edge_iterator;
		out_edge_iterator current_e, last_e;
		for (tie(current_e, last_e) = out_edges(*current_v, in_graph); current_e != last_e; ++current_e) {

			// Run the dfs to get the path
			dfs.run(*current_e);

			// Erase the path from G
			//dfs.delete_path();

			// Remove self-intersections...
			bypass_intersections<path> bi(dfs.get_path());
			bi.run(eps);

			// Run Ramer-Douglas-Peucker algorithm and insert the simplified path into G
			ramer_douglas_peucker<path> rdp(bi.get_simplified_path());
			rdp.run(eps);

			// Insert the simplified path
			insert_path(rdp.get_simplified_path(), vertex_map[dfs.get_source_node()], vertex_map[dfs.get_target_node()], out_graph);
		}
	}

    //fix edge ids
    int edge_id = 0;
  
    typedef typename graph_traits<graphType>::edge_descriptor edge_descriptor;
    edge_descriptor e;
    BGL_FORALL_EDGES_T(e, out_graph, graphType){
      out_graph[e].edge_id = edge_id++;
    }

}



// This algorithm decides if a nodes is important by checking its degree and
// if its in_degree = 1 and out_degree = 1 checks whether the incident edges
// have the same properties.
//GEMSA TODO: The stated criteria are not always the best choice. We should
// think about including sharp turns, etc.
template<typename graphType>
bool is_important_node(typename graphType::vertex_descriptor v, graphType &graph) {
	if (in_degree(v, graph) != 1)
		return true;
	if (out_degree(v, graph) != 1)
		return true;

	typename graphType::edge_descriptor in_edge = get_first_in_edge(v, graph);
	typename graphType::edge_descriptor out_edge = get_first_out_edge(v, graph);
	if (graph[in_edge].category != graph[out_edge].category || graph[in_edge].name != graph[out_edge].name)
		return true;

	return false;
}

}

#endif /* SIMPLIFICATION_H_ */
