/******************************************************************************
* path.h
*
* RouteSketch
*
******************************************************************************
* Copyright (C) 2013 Andreas Gemsa <gemsa@kit.edu>, Thomas Pajor
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation, either version 3 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/
/*
 * path.h
 *
 *  Created on: Jul 1, 2010
 *      Author: pajor
 */

#ifndef PATH_H_
#define PATH_H_

#include "../geometry/point.h"

#include <vector>
#include <algorithm>


namespace schematization {

struct path_node {
	point coord;

	bool to_erase; // indicates whether this node may be removed from the path.

	path_node() : coord(0.0, 0.0), to_erase(false) {}
	path_node(const point &c) : coord(c), to_erase(false) {}
};


// Ostream function for the path node
std::ostream &operator<<(std::ostream &os, const path_node &n) {
	const char tab('\t');
	return os << n.coord.x << tab << n.coord.y;
}




class path {
public:

	signed char category;
	std::string name;

	bool context;
public:

	path(signed char c, std::string &n) : category(c), name(n) {}
	path() {}

	// Clear path
	void clear() {
	  nodes.clear();
	}

	// Appends a new node (specifying the new edge properties)
	// returns: new node_id
	size_t push_node(const point p) {
	  path_node new_node(p);
	  return push_node(new_node);
	}
	size_t push_node(const path_node &n) {
	  nodes.push_back(n);
	  return nodes.size() - 1;
	}

	// returns: length of path in number of nodes (beware of the 'alleenproblem')
	size_t size() const {
	  return nodes.size();
	}

	// Reverse the path (e.g., useful if the path has been created through back-traversing a shortest-path-tree)
	void reverse() {
	  std::reverse(nodes.begin(), nodes.end());
	}

	// Access nodes on the path.
	path_node &operator[](const size_t n) {
	  return nodes[n];
	}
	const path_node &operator[](const size_t n) const {
	  return nodes[n];
	}

	// This method erases all nodes from the path whose nodes have the to_erase flag set.
	void reduce() {
	  nodes.erase(remove_if(nodes.begin(), nodes.end(), is_erasable), nodes.end());
	}

protected:

	// Check if the node is erasable
	inline static bool is_erasable(const path_node &n) {
		return n.to_erase;
	}

	// These are the nodes on the path.
	vector<path_node> nodes;
};


// Operator <<: output the path in human readable form.
std::ostream &operator<<(std::ostream &os, const path &p) {
	os << "% This is a path with " << p.size() << " nodes." << endl
	   << "x\ty\tthickn\tcol_r\tcol_g\tcol_b\tcol_alpha" << endl;
	for (std::size_t n = 0; n < p.size(); ++n) {
		os << p[n] << std::endl;
	}
	return os;
}


}

#endif /* PATH_H_ */
