/******************************************************************************
* schemMIP.h
*
* RouteSketch
*
******************************************************************************
* Copyright (C) 2013 Andreas Gemsa <gemsa@kit.edu>, Thomas Pajor
*
* This program is free software: you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation, either version 3 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*****************************************************************************/
/*
 * schemMIP.h
 *
 *  Created on: Jul 6, 2010
 *      Author: pajor
 */

#ifndef SCHEMMIP_H_
#define SCHEMMIP_H_

#include "../dataStructures/graph.h"
#include "../algorithms/orthogonal_order.h"


#include <boost/foreach.hpp>
#include <boost/multi_array.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/iteration_macros.hpp>
#include <map>
#include <queue>
#include <algorithm>

#include "gurobi_c++.h"


namespace schematization {


  template<typename graphType>
    class schemMIP {

  public:

    typedef typename graphType::vertex_descriptor vertex_descriptor;
    typedef typename graphType::edge_descriptor edge_descriptor;
    typedef typename graph_traits<graphType>::vertex_iterator vertex_iterator;
    typedef typename graph_traits<graphType>::edge_iterator edge_iterator;

  public:

  schemMIP(GRBEnv &env, graphType &in_g, const angleType &a, int orthogonal_order_ = 1) :
    graph(in_g),
      angle(a),
      angle_rad(tools::deg_to_rad(angle)),
      num_nodes(num_vertices(graph)),
      num_arcs(num_edges(graph)),
      num_sectors(static_cast<size_t>(360.0 / angle)),
      x_axis(0),
      y_axis(num_sectors / 4),
      canvas_extent_x(paper_size_w * 10 - 2*path_margin),
      canvas_extent_y(paper_size_h * 10 - 2*path_margin),
      M(tools::dist_2p(point(0,0), point(canvas_extent_x, canvas_extent_y))),
      gurobi_env(env),
      gurobi_model(gurobi_env),
      do_output(true),
      num_iterations(0),
      orthogonal_order(orthogonal_order_)  {

      }


    ~schemMIP() {
    }

    // Disable output
    inline void disable_output() {
      do_output = false;
    }


    /*
     * This method builds the linear programming model given some graph
     */
    void solve(graphType &out_graph) {
      if (do_output) cout << "Graph has " << num_nodes << " nodes and " << num_arcs << " edges" << endl;
      // Build the model
      Timer tau;
      if (do_output) cout << "Building model... " << flush;

      build_model();
      if (do_output) cout << "done (" << tau.elapsed() << " ms)" << endl;

      // Optimize the model until it is planar
      if (do_output) cout << "Optimizing..." << flush;
      tau.restart();
      num_iterations = 0;
      size_t num_planarity_constraints = 0;
      do {
	++num_iterations;
	gurobi_model.optimize();

	num_planarity_constraints = add_planarity_constraints();
	if (do_output && num_planarity_constraints > 0) cout << ".(" << num_planarity_constraints << ")." << flush;
      } while (num_planarity_constraints > 0 && num_iterations < 10);
      if (do_output) cout << " done (" << num_iterations << " iterations in " << tau.elapsed() << " ms)" << endl;
		
      //write model
      //dump_model("ausgabe.lp");

      if (is_optimal())
	{

	  //write solution 
	  //dump_model("ausgabe.sol");

	  // Extract the graph
	  if (do_output) cout << "Extracting graph... " << flush;
	  tau.restart();
	  extract_graph(out_graph);
	  if (do_output) cout << "done (" << tau.elapsed() << " ms)" << endl;
	  //cout << "dev cost: " << v_dev_cost.get(GRB_DoubleAttr_X) << endl;
	  edge_descriptor e;

	}

      else
	{
	  //gurobi_model.computeIIS();
	  //dump_model("ausgabe.lp");
	  cout << endl << "no solution found! GRB status: " << gurobi_model.get(GRB_IntAttr_Status) << endl << endl;
	}
    }


    /*
     * See if the model was proven to be infeasible
     */
    inline bool is_infeasible() const {
      return gurobi_model.get(GRB_IntAttr_Status) == GRB_INFEASIBLE || gurobi_model.get(GRB_IntAttr_Status) == GRB_INF_OR_UNBD;
    }

    inline bool is_solved() const {
      return !is_optimal();
    }

    inline bool is_optimal() const {
      return gurobi_model.get(GRB_IntAttr_Status) == GRB_OPTIMAL;
    }

    /*
     * Retrieves the number of iterations of the last solution
     */
    inline size_t get_num_iterations() const {
      return num_iterations;
    }

    /*
     * This dumps the current model to some file
     */
    void dump_model(const string &filename) {
      gurobi_model.write(filename);
    }



  protected:




    /*
     * Builds the gurobi model for the given input path and some angle
     */
    bool build_model() {		
		

      try {
	// Initialize the variable array

	create_vars();

	// Create Variables: Coordinates
	for (size_t n = 0; n < num_nodes; ++n) {
	  v_coord[n][x_axis] = gurobi_model.addVar(-canvas_extent_x / 2.0, canvas_extent_x / 2.0, 0.0, GRB_CONTINUOUS, "v_coord[" + boost::lexical_cast<string>(n) + "][" + boost::lexical_cast<string>(x_axis) + "]");
	  v_coord[n][y_axis] = gurobi_model.addVar(-canvas_extent_y / 2.0, canvas_extent_y / 2.0, 0.0, GRB_CONTINUOUS, "v_coord[" + boost::lexical_cast<string>(n) + "][" + boost::lexical_cast<string>(y_axis) + "]");

	  for (size_t a = 0; a < num_sectors; ++a) {
	    if (a == x_axis || a == y_axis) continue;
	    v_coord[n][a] = gurobi_model.addVar(-M, M, 0.0, GRB_CONTINUOUS, "v_coord[" + boost::lexical_cast<string>(n) + "][" + boost::lexical_cast<string>(a) + "]");
	  }
	}

	// Create Variables: decision variables for the admissible angles

	for (size_t n = 0; n < num_arcs; ++n) {
	  for (size_t a = 0; a < num_sectors; ++a) {
	    v_dir[n][a] = gurobi_model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "v_dir[" + boost::lexical_cast<string>(n) + "][" + boost::lexical_cast<string>(a) + "]");
	  }
	}

	// Create Variables: Distances
	for (size_t n = 0; n < num_arcs; ++n) {
	  v_dist[n] = gurobi_model.addVar(0.0, M, 1.0/(M), GRB_CONTINUOUS, "v_dist[" + boost::lexical_cast<string>(n) + "]");
	}

	// Create Variables: Planarity
	for (int n = 0; n < num_arcs; ++n) {
	  for (int m = 0; m < num_arcs; ++m) {
	    for (size_t a = 0; a < num_sectors; ++a) {
	      v_pos[n][m][a] = gurobi_model.addVar(0.0, 1.0, 0.0, GRB_BINARY, "v_pos[" + boost::lexical_cast<string>(n) + "][" + boost::lexical_cast<string>(m) + "][" + boost::lexical_cast<string>(a) + "]");
	    }
	  }
	}


	// Create Variable: Sector deviation minimization
	v_dev_cost = gurobi_model.addVar(0.0, double(num_sectors*num_nodes)/2.0, 1.0, GRB_CONTINUOUS, "v_dev_cost");

	gurobi_model.update();

	// Constraints: Induce coordinate axes
	for (size_t n = 0; n < num_nodes; ++n) {
	  for (size_t a = 0; a < num_sectors; ++a) {
	    if (a == x_axis || a == y_axis)
	      continue;
	    gurobi_model.addConstr(v_coord[n][a] == v_coord[n][x_axis] * tools::round(std::cos(double(a) * angle_rad), 6)
				   + v_coord[n][y_axis] * tools::round(std::sin(double(a) * angle_rad), 6), "ica" + boost::lexical_cast<string>(n) + ":" + boost::lexical_cast<string>(a) );
	  }
	}


	if (orthogonal_order == 0){	
	  cout << "orthogonal order is *not* considered!" << endl;
	}
	else if	(orthogonal_order == 1){
				
	  // Constraints: Orthogonal x-Order.
	  vector<size_t> x_order = get_orthogonal_order(X_ORDER, graph);
	  for (size_t n = 1; n < x_order.size(); ++n) {			
	    setOrthogonalXOrder(x_order[n-1], x_order[n]);
	  }

	  // Constraints: Orthogonal y-Order.
	  vector<size_t> y_order = get_orthogonal_order(Y_ORDER, graph);
	  for (size_t n = 1; n < y_order.size(); ++n) {				
	    setOrthogonalYOrder(y_order[n-1], y_order[n]);
	  }

	}

	// Constraints: Edges must conform to one of the coordinate axes, and also satisfy some min_length
	GRBLinExpr dev_expr;
	edge_iterator current_edge, last_edge;
	size_t current_edge_index(0);
	for (tie(current_edge, last_edge) = edges(graph); current_edge != last_edge; ++current_edge) {
	  // Source and target nodes of the edge
	  vertex_descriptor u(source(*current_edge, graph));
	  vertex_descriptor v(target(*current_edge, graph));
	  current_edge_index = graph[*current_edge].edge_id;

	  GRBLinExpr sum_expr = 0;
	  int first_axis = get_edge_quadrant(*current_edge, graph) * y_axis;
	  int last_axis  = first_axis + y_axis;
	  GRBLinExpr sum_expr2 = 0;
                
	  for (int axis = first_axis; axis <= last_axis; ++axis) {
	    int a = axis % num_sectors;
	    // In order to conform to dir[a], the orthogonal axis coordinate has to be equal for that edge
	    gurobi_model.addConstr(v_coord[v][(a + y_axis) % num_sectors] - v_coord[u][(a + y_axis) % num_sectors] + M * v_dir[current_edge_index][a] <= M, "conform1");
	    gurobi_model.addConstr(v_coord[v][(a + y_axis + num_sectors/2) % num_sectors] - v_coord[u][(a + y_axis + num_sectors/2) % num_sectors] + M * v_dir[current_edge_index][a] <= M, "conform2");
                    
	    // Compute minimum distance for this edge
	    double sc_x = graph[v].coord.x - graph[u].coord.x;
	    double sc_y = graph[v].coord.y - graph[u].coord.y;

	    double min_dist_n = category_to_min_length[graph[*current_edge].category];

	    gurobi_model.addConstr(v_coord[v][a] - v_coord[u][a] >= - M + M * v_dir[current_edge_index][a] + v_dist[current_edge_index], " lower bound dist"); 
	    gurobi_model.addConstr(v_coord[v][a] - v_coord[u][a] >= - M + M * v_dir[current_edge_index][a] + min_dist_n, " lower bound dist");
	    gurobi_model.addConstr(v_coord[v][a] - v_coord[u][a] <=   M - M * v_dir[current_edge_index][a] + v_dist[current_edge_index], " upper bound dist");

	    int sector(get_edge_sector(*current_edge, angle, graph));
					
	    if (sector != a){
	      int q = get_edge_quadrant(*current_edge, graph);
	      int dist_sector = a + ( (q == 3 && a == 0) ? num_sectors : 0);	
	      int my_sector = (q == 3 && sector == 0) ? num_sectors : sector;

	      dev_expr += v_dir[current_edge_index][a] * ( std::abs(my_sector- dist_sector) );
	    }
	    // Construct sum expression
	    sum_expr += v_dir[current_edge_index][a];
	  }
	  //Exactly one of the directions in the appropriate quadrant is allowed
	  gurobi_model.addConstr(sum_expr == 1.0, "only one direction is allowed");

	  //No direction which forces an edge outside "its" quadrant is allowed (i.e., the sum of the binary values has to be 0)
	  sum_expr = 0;
	  for (int a = 0; a < num_sectors; a++)
	    {	
	      if (a >= first_axis &&  a <= last_axis || (last_axis == num_sectors && a == 0)) //valid edge directions.
		continue;

	      sum_expr += v_dir[current_edge_index][a];
	    }
	  gurobi_model.addConstr(sum_expr == 0.0, "only one direction is allowed#");

	}

	// Constraint: define deviation cost
	gurobi_model.addConstr(v_dev_cost == dev_expr, "deviation cost");

	/*
	  
	  Loops through all vertices and compares pairwise all incident edges (both in- and outgoing)
	  Ensures that constraints are in place that keep the cyclic order intact

	*/	

	vertex_iterator v_current_it, v_last_it;
	typedef typename graph_traits<graphType>::out_edge_iterator out_edge_iterator;
	typedef typename graph_traits<graphType>::in_edge_iterator in_edge_iterator;
	//loop through all vertices
	for (tie(v_current_it, v_last_it) = vertices(graph); v_current_it != v_last_it; ++v_current_it) {
	  point source_point = point(graph[*v_current_it].coord.x, graph[*v_current_it].coord.y);

	  out_edge_iterator out_edge_it_outer, out_edge_it_last, out_edge_it_inner;
	  in_edge_iterator   in_edge_it_outer,  in_edge_it_last,  in_edge_it_inner;

	  //compare outgoing edges with remaining outgoing and all incoming edges
	  for (tie(out_edge_it_outer, out_edge_it_last) = out_edges(*v_current_it, graph); out_edge_it_outer != out_edge_it_last; ++out_edge_it_outer) {
	    vertex_descriptor v = target(*out_edge_it_outer, graph);
	    int first_edge_index = graph[*out_edge_it_outer].edge_id;

	    line l1(source_point,  point(graph[v].coord.x, graph[v].coord.y));
	    int quadrant1 = tools::get_quadrant(l1) ;

	    //outgoing edges
	    for (out_edge_it_inner = out_edge_it_outer + 1; out_edge_it_inner != out_edge_it_last; ++out_edge_it_inner) {

	      vertex_descriptor v_target2 = target(*out_edge_it_inner, graph);
	      int second_edge_index = graph[*out_edge_it_inner].edge_id;

	      line l2(source_point, point(graph[v_target2].coord.x, graph[v_target2].coord.y));
	      int quadrant2 = tools::get_quadrant(l2);
						
	      if (quadrant1 == quadrant2) {
		  add_cyclic_order_constraint(first_edge_index, second_edge_index, quadrant1, tools::rad_to_deg(tools::get_angle(l1)), tools::rad_to_deg(tools::get_angle(l2)), false, false );
	      }
	      else if (abs(quadrant1 - quadrant2) != 2) { //both edges are in neighbouring quadrants
		add_consistency_constraint(quadrant1, quadrant2, first_edge_index, second_edge_index, false, false);
	      }
	    }
					
	    //incoming edges
	    for (tie(in_edge_it_inner, in_edge_it_last) = in_edges(*v_current_it, graph); in_edge_it_inner != in_edge_it_last; ++in_edge_it_inner) {
	      vertex_descriptor v_target2 = source(*in_edge_it_inner, graph);
	      int second_edge_index = graph[*in_edge_it_inner].edge_id;
	      
	      line l2(source_point, graph[v_target2].coord);
	      int quadrant2 = tools::get_quadrant(l2);

		if (quadrant1 == quadrant2)  {
		  add_cyclic_order_constraint(first_edge_index, second_edge_index, quadrant1, tools::rad_to_deg(tools::get_angle(l1)), tools::rad_to_deg(tools::get_angle(l2)), false, true );
		}
		else if (abs(quadrant1 - quadrant2) != 2) { //both edges are in neighbouring quadrants
		  add_consistency_constraint(quadrant1, quadrant2, first_edge_index, second_edge_index, false, true);
		}
	    }

	  }

	  //compare all incoming edges with each other
	  for (tie(in_edge_it_outer, in_edge_it_last) = in_edges(*v_current_it, graph); in_edge_it_outer != in_edge_it_last; ++in_edge_it_outer) 
	    {
	      vertex_descriptor v = source(*in_edge_it_outer, graph);
	      int first_edge_index = graph[*in_edge_it_outer].edge_id;

	      line l1(source_point,  point(graph[v].coord.x, graph[v].coord.y));			
	      int quadrant1 = tools::get_quadrant(l1);

	      //incoming edges
	      for (in_edge_it_inner= in_edge_it_outer + 1; in_edge_it_inner!= in_edge_it_last; ++in_edge_it_inner) 
		{
		  vertex_descriptor v_target2 = source(*in_edge_it_inner, graph);
		  line l2(source_point, point(graph[v_target2].coord.x, graph[v_target2].coord.y));

		  //						int second_edge_index = extGraphProps.getEdgeId(*in_edge_it_inner, graph);				
		  int second_edge_index = graph[*in_edge_it_inner].edge_id;
		  int quadrant2 = tools::get_quadrant(l2);
						
		  if (quadrant1 == quadrant2)
		    {
		      add_cyclic_order_constraint(first_edge_index, second_edge_index, quadrant1, tools::rad_to_deg(tools::get_angle(l1)), tools::rad_to_deg(tools::get_angle(l2)), true, true );
		    }
		  else if (abs(quadrant1 - quadrant2) != 2) //both edges are in neighbouring quadrants
		    {
		      add_consistency_constraint(quadrant1, quadrant2, first_edge_index, second_edge_index, true, true);
		    }

						
		}
	    }
	}
      } catch (GRBException e) {
	cout << "Error code = " << e.getErrorCode() << endl;
	cout << e.getMessage() << endl;
	return false;
      } catch (...) {
	cout << "Exception during optimization" << endl;
	return false;
      }
      return true;
    }

    /*
      Adds constraints so that two edges incident to the same node, embedded in two adjacent quadrants
      cannot both be embedded at the shared sector
    */

    bool add_consistency_constraint(const int quadrant1, const int quadrant2, const int edge_index1, const int edge_index2, bool in1, bool in2){
		
      int q_max = std::max(quadrant1, quadrant2);
      int q_min = std::min(quadrant1, quadrant2);

      int conflict_sector = (q_max == 3 && q_min == 0) ? 8 			: q_max * num_sectors/4;
      int first_sector =    (q_max == 3 && q_min == 0) ? 3 * num_sectors/4 	: q_min * num_sectors/4;
	
      int e1 = (quadrant1 < quadrant2 || (quadrant1 == 3 && quadrant2 == 0 )) ? edge_index1 : edge_index2;
      int e2 = (quadrant1 < quadrant2 || (quadrant1 == 3 && quadrant2 == 0 )) ? edge_index2 : edge_index1;

      int add_smaller = (quadrant1 < quadrant2 || (quadrant1 == 3 && quadrant2 == 0 )) ? (in1 ? num_sectors / 2: 0) : (in2 ? num_sectors / 2: 0);
      int add_larger =  (quadrant1 < quadrant2 || (quadrant1 == 3 && quadrant2 == 0 )) ? (in2 ? num_sectors / 2: 0) : (in1 ? num_sectors / 2: 0);

      GRBLinExpr sum_expr = 0;

      for (int l = first_sector; l < conflict_sector; l++){
	//cout << l << " " << first_sector << " " << conflict_sector << " " << q_max << " " << q_min << " " << quadrant1 << " " << quadrant2 << endl;
	sum_expr += v_dir[e1][ (l + add_smaller) % num_sectors];
      }

      gurobi_model.addConstr(sum_expr >= v_dir[e2][ (conflict_sector + add_larger) % num_sectors], "consistency constraints");	

      return true;
    }

    bool add_cyclic_order_constraint(const int edge_index1, const int edge_index2, int quadrant1, const double winkel_l1, const double winkel_l2, bool in1, bool in2){

      int num_sectors_per_quadrant = num_sectors / 4 + 1;
		
      int first_sector = quadrant1 * num_sectors / 4;

      int last_sector = (quadrant1 + 1) * num_sectors / 4;

      int smaller_e = (winkel_l1 <= winkel_l2) ? edge_index1 : edge_index2;
      int larger_e =  (winkel_l1 <= winkel_l2) ? edge_index2 : edge_index1;	
	
      int add_smaller = (winkel_l1 <= winkel_l2) ? (in1 ? num_sectors / 2: 0) : (in2 ? num_sectors / 2: 0);
      int add_larger = (winkel_l1 <= winkel_l2) ? (in2 ? num_sectors / 2: 0) : (in1 ? num_sectors / 2: 0);


      for (int k(first_sector); k < last_sector; k++){
	GRBLinExpr sum_expr = 0;

	for (int j(k+1); j <= last_sector; j++){
	  sum_expr += v_dir[larger_e][ (j + add_larger) % num_sectors];
	}
	
	gurobi_model.addConstr(sum_expr >= v_dir[smaller_e][(k + add_smaller) % num_sectors], "cyclic order");

		
      }
      gurobi_model.addConstr(v_dir[smaller_e][(add_smaller+last_sector) % num_sectors] == 0, "cyclic order");


    }

    /*
     * Adds planarity constraints to the model for conflicting edges
     * This tests each edge pairs for conflicts (intersections), adds
     * the constraints and returns the number of constraints added.
     *
     * If the number of constraints returned is zero, no changes were
     * made to the model.
     */
    size_t add_planarity_constraints() {
      unsigned int num_planarity_constraints(0);

      // Don't try to do stuff if the model is infeasible!
      if (is_infeasible())
	return 0;

      try {
	// Try all pairs of edges
	edge_iterator e1,e1_end,e2,e2_end;
	size_t n(0), m(0);
	for (tie(e1,e1_end) = edges(graph), n = 0; e1 != e1_end; ++e1, ++n) {
	  for (tie(e2,e2_end) = edges(graph), m = 0; e2 != e2_end; ++e2, ++m) {

	    if (!has_planarity_conflict(*e1,*e2))
	      continue;
						

	    // Build expression saying, that at least one pos has to be true
	    GRBLinExpr sum_expr = 0;
	    for (size_t a = 0; a < num_sectors; ++a)
	      sum_expr += v_pos[n][m][a];
	    gurobi_model.addConstr(sum_expr >= 1.0, "");

	    // For every direction, model position constraints, that are satisfied if both vertices of edge m are
	    // beyond vertices of edge n
	    for (size_t a = 0; a < num_sectors; ++a) {
	      vertex_descriptor u1(source(*e1,graph)), v1(target(*e1,graph)), u2(source(*e2,graph)), v2(target(*e2,graph));
	      // Direction e1 -> e2
	      gurobi_model.addConstr(v_coord[u1][a] - v_coord[u2][a] + M * v_pos[n][m][a] <= M - min_edge_dist);
	      gurobi_model.addConstr(v_coord[u1][a] - v_coord[v2][a] + M * v_pos[n][m][a] <= M - min_edge_dist);
	      gurobi_model.addConstr(v_coord[v1][a] - v_coord[u2][a] + M * v_pos[n][m][a] <= M - min_edge_dist);
	      gurobi_model.addConstr(v_coord[v1][a] - v_coord[v2][a] + M * v_pos[n][m][a] <= M - min_edge_dist);
	    }
	    ++num_planarity_constraints;
	  }
	}
      } catch (GRBException e) {
	cout << "Error code = " << e.getErrorCode() << endl;
	cout << e.getMessage() << endl;
      } catch (...) {
	cout << "Exception during optimization" << endl;
      }

      return num_planarity_constraints;
    }


    inline void setOrthogonalXOrder(size_t index1, size_t index2){
      // If x-coordinates are equal, then use "=" in constraint,
      // otherwise use "<="
      //cout << "x-order: " << index1 << " [] " << index2 << endl;
      if (graph[index1].coord.x == graph[index2].coord.x) {
	gurobi_model.addConstr(v_coord[index1][x_axis] == v_coord[index2][x_axis], "orthogonal x-order");
      } else {
	gurobi_model.addConstr(v_coord[index1][x_axis] <= v_coord[index2][x_axis], "orthogonal x-order");
      }		
    }

    inline void setOrthogonalYOrder(size_t index1, size_t index2){
      // If y-coordinates are equal, then use "=" in constraint,
      // otherwise use "<="
      if (graph[index1].coord.y == graph[index2].coord.y) {
	gurobi_model.addConstr(v_coord[index1][y_axis] == v_coord[index2][y_axis], "orthogonal y-order");
      } else {
	gurobi_model.addConstr(v_coord[index1][y_axis] <= v_coord[index2][y_axis], "orthogonal y-order");
      }		
    }


    /*
     * This tests whether two edges have a planarity conflict in the given solution of the model
     */
    bool has_planarity_conflict(const edge_descriptor e1, const edge_descriptor e2) {
	
      // Edges which are adjacent (or the same) are never conflicting...
      if (e1 == e2 || is_adjacent(e1, e2, graph))
	return false;
			
		
      // We say e1 = (u1, v1) and e2 = (u2, v2)
      // Thus we have [(xu1,yu1),(xv1,yv1)] and [(xu2,yu2),(xv2,yv2)]
      vertex_descriptor u1(source(e1, graph)), v1(target(e1, graph)), u2(source(e2, graph)), v2(target(e2, graph));

      const double xu1 = v_coord[u1][x_axis].get(GRB_DoubleAttr_X);
      const double yu1 = v_coord[u1][y_axis].get(GRB_DoubleAttr_X);
      const double xv1 = v_coord[v1][x_axis].get(GRB_DoubleAttr_X);
      const double yv1 = v_coord[v1][y_axis].get(GRB_DoubleAttr_X);
      const double xu2 = v_coord[u2][x_axis].get(GRB_DoubleAttr_X);
      const double yu2 = v_coord[u2][y_axis].get(GRB_DoubleAttr_X);
      const double xv2 = v_coord[v2][x_axis].get(GRB_DoubleAttr_X);
      const double yv2 = v_coord[v2][y_axis].get(GRB_DoubleAttr_X);

      //
      //cout << "lines:" << endl;
      //cout << xu1 << ", " << yu1 << " - " << xv1 << ", " << yv1 << endl;
      //cout << xu2 << ", " << yu2 << " - " << xv2 << ", " << yv2 << endl;
      //
    
      const double A1 = yv1 - yu1;
      const double B1 = xu1 - xv1;
      const double C1 = A1*xu1 + B1*yu1;
      const double A2 = yv2 - yu2;
      const double B2 = xu2 - xv2;
      const double C2 = A2*xu2 + B2*yu2;

      const double det = A1*B2 - A2*B1;

      bool intersect = true;

      // Lines are parallel, so no overlap [AG: ... that is not really correct.. is it?]
      if (det == 0)
	{
		    
	  intersect = false;
	}
      else {
	// If line Segments are not parallel then test for intersection...
	// Compute intersection point
	double x = (B2*C1 - B1*C2)/det;
	double y = (A1*C2 - A2*C1)/det;
			
	//cout << "intersection point: " << x << ", " << y << endl;
	//char c;
	//cin >> c;
	// Test whether intersection point is contained on the line segments
	double epsilon = 0.1; //circumenvents problems with float point arithmetic
	if (! (std::min(xu1,xv1) - epsilon <= x && x <= std::max(xu1,xv1) + epsilon) )
	  intersect = false;
	if (! (std::min(yu1,yv1) - epsilon <= y && y <= std::max(yu1,yv1) + epsilon) )
	  intersect = false;
	if (! (std::min(xu2,xv2) - epsilon <= x && x <= std::max(xu2,xv2) + epsilon) )
	  intersect = false;
	if (! (std::min(yu2,yv2) - epsilon <= y && y <= std::max(yu2,yv2) + epsilon) )
	  intersect = false;
      }
      if (intersect)
	return true;
		

      //		return false;

      // Test for distance between the two line segments
      // We know here, that they do not overlap
      // If dist < min_dist_edge then return true as well (since they need to be moved apart)
		
      double dist = std::min(get_dist_node_to_edge(e1,u2),
			     std::min(get_dist_node_to_edge(e1,v2),
				      std::min(get_dist_node_to_edge(e2, u1),
					       get_dist_node_to_edge(e2, v1))));

      return (dist < min_edge_dist);
    }



    /*
     * This computes the distance between an edge (induced by its head node) and a node
     * in the solution of the MIP
     */
    inline double get_dist_node_to_edge(const edge_descriptor e, size_t n) {
      const vertex_descriptor u(source(e, graph)), v(target(e, graph));
      const double xu = v_coord[u][x_axis].get(GRB_DoubleAttr_X);
      const double yu = v_coord[u][y_axis].get(GRB_DoubleAttr_X);
      const double xv = v_coord[v][x_axis].get(GRB_DoubleAttr_X);
      const double yv = v_coord[v][y_axis].get(GRB_DoubleAttr_X);
      const double xn = v_coord[n][x_axis].get(GRB_DoubleAttr_X);
      const double yn = v_coord[n][y_axis].get(GRB_DoubleAttr_X);

      const double len_e = tools::sqr(yv-yu) + tools::sqr(xv-xu);
      assert(len_e > 0);
      const double t = ((xn-xu)*(xv-xu) + (yn-yu)*(yv-yu)) / len_e;
      if (t < 0.0)
	return std::sqrt(tools::sqr(xn-xu) + tools::sqr(yn-yu));
      if (t > 1.0)
	return std::sqrt(tools::sqr(xn-xv) + tools::sqr(yn-yv));

      return std::fabs((xv-xu)*(yu-yn) - (xu-xn)*(yv-yu)) / std::sqrt(tools::sqr(xv-xu) + tools::sqr(yv-yu));
    }


    /*
     * This creates the gurobi variables
     */
    inline void create_vars() {
      v_coord.resize(boost::extents[num_nodes][num_sectors]);
      v_dir.resize(boost::extents[num_arcs][num_sectors]);
      v_dist.resize(boost::extents[num_arcs]);
      v_pos.resize(boost::extents[num_arcs][num_arcs][num_sectors]);
    }



    /*
     * This extracts a path from a solved gurobi model
     */
    void extract_graph(graphType &out_graph) {
      try {
	out_graph = graph;
	for (size_t n = 0; n < num_nodes; ++n) {
	  out_graph[n].coord.x = v_coord[n][x_axis].get(GRB_DoubleAttr_X);
	  out_graph[n].coord.y = v_coord[n][y_axis].get(GRB_DoubleAttr_X);
	}
      } catch (GRBException e) {
	cout << "Error code = " << e.getErrorCode() << endl;
	cout << e.getMessage() << endl;
      } catch (...) {
	cout << "Exception during optimization" << endl;
      }
    }




  private:
    // Cooordinates for the points
    boost::multi_array<GRBVar, 2> v_coord; //[num_nodes][num_sectors];
    // Decision variables for the angles
    boost::multi_array<GRBVar, 2> v_dir; //[num_nodes][num_sectors];
    // Distance variables
    boost::multi_array<GRBVar, 1> v_dist; //[num_nodes];
    // Planarity variables
    boost::multi_array<GRBVar, 3> v_pos; //[num_nodes][num_nodes][num_sectors];

    // This holds the sum for the sector deviation minimization
    GRBVar v_dev_cost;

    // Stuff belonging to gurobi (environment, model, ...)
    GRBEnv gurobi_env;
    GRBModel gurobi_model;

    // This is the input graph, that should be schematized by this model
    graphType &graph;

    // This is the (current) angle we schematize
    const angleType angle;

    // Some constants induced by the paths and the angles
    size_t num_nodes;
    size_t num_arcs;
    const size_t num_sectors;
    const double angle_rad;
    const size_t x_axis;
    const size_t y_axis;
    const double canvas_extent_x;
    const double canvas_extent_y;

    // Some size constants
    const double M;

    vector<int>pref_edge_dir;
	
    // This indicates whether we have output
    bool do_output;

    // The number of iterations we did until we found the solution
    size_t num_iterations;

    //
    bool maintain_length_order;
    int orthogonal_order;
    bool alt_route_length_deviation;
	
};

}

#endif /* SCHEMMIP_H_ */
